/*
 * Copyright 2018 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.javadoc;

import com.google.errorprone.CompilationTestHelper;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/** Unit tests for {@link InheritDoc} bug pattern. */
@RunWith(JUnit4.class)
public final class InheritDocTest {
  private final CompilationTestHelper helper =
      CompilationTestHelper.newInstance(InheritDoc.class, getClass());

  @Test
  public void nonOverridingMethod() {
    helper
        .addSourceLines(
            "Test.java",
            "class Test {",
            "  // BUG: Diagnostic contains: InheritDoc",
            "  /** {@inheritDoc} */",
            "  void test() {}",
            "}")
        .doTest();
  }

  @Test
  public void overridingMethod() {
    helper
        .addSourceLines(
            "Test.java",
            "class Test {",
            "  /** {@inheritDoc} */",
            "  @Override public boolean equals(Object o) { return true; }",
            "}")
        .doTest();
  }

  @Test
  public void nonOverridingClass() {
    helper
        .addSourceLines(
            "Test.java",
            "// BUG: Diagnostic contains: InheritDoc",
            "/** {@inheritDoc} */",
            "class Test {}")
        .doTest();
  }

  @Test
  public void overridingClass() {
    helper
        .addSourceLines(
            "Test.java", //
            "/** {@inheritDoc} */",
            "class Test extends java.util.ArrayList {}")
        .doTest();
  }

  @Test
  public void variable() {
    helper
        .addSourceLines(
            "Test.java",
            "class Test {",
            "  // BUG: Diagnostic contains: InheritDoc",
            "  /** {@inheritDoc} */",
            "  private static final int a = 1;",
            "}")
        .doTest();
  }
}
