/*
 * Copyright 2017 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns;

import com.google.errorprone.CompilationTestHelper;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/** {@link ShortCircuitBoolean}Test */
@RunWith(JUnit4.class)
public class ShortCircuitBooleanTest {

  @Test
  public void positive() {
    CompilationTestHelper.newInstance(ShortCircuitBoolean.class, getClass())
        .addSourceLines(
            "Test.java",
            "class Test {",
            "  void f(boolean b1, boolean b2, boolean b3, int i1, int i2) {",
            "    // BUG: Diagnostic contains: b1 && b2",
            "    boolean r = b1 & b2;",
            "    // BUG: Diagnostic contains: b1 || b2",
            "    r = b1 | b2;",
            "    // BUG: Diagnostic contains: b1 || !b2 || b3 && b1 && b2",
            "    r = b1 | !b2 | b3 & b1 & b2;",
            "    // BUG: Diagnostic contains: b1 || (b2 && b3) && (b1 || b2)",
            "    r = b1 | (b2 & b3) & (b1 | b2);",
            "    // BUG: Diagnostic contains: b1 || (b2 && b3)",
            "    r = b1 | (b2 & b3);",
            "    // BUG: Diagnostic contains: (b1 || b2) && b3",
            "    r = (b1 | b2) & b3;",
            "    // BUG: Diagnostic contains: b1 || b3 && b3",
            "    r = b1 | b3 && b3;",
            "    // BUG: Diagnostic contains: b1 || b2 == b3",
            "    r = b1 | b2 == b3;",
            "    // BUG: Diagnostic contains: (b1 || b2) != b3",
            "    r = (b1 | b2) != b3;",
            "    int i = i1 | i2;",
            "    i = i1 & i2;",
            "    i = i1 & i2 | i1;",
            "  }",
            "}")
        .doTest();
  }
}
