/*
 * Copyright 2012 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.nullness.testdata;

import static com.google.common.base.Preconditions.checkNotNull;

import com.google.common.base.Preconditions;

public class UnnecessaryCheckNotNullPrimitiveNegativeCases {
  public void test() {
    Object obj1 = new Object();

    Preconditions.checkNotNull(obj1);
    checkNotNull(obj1);
    Preconditions.checkNotNull(obj1, "obj1 should not be null");
    Preconditions.checkNotNull(obj1, "%s should not be null", "obj1");
    Preconditions.checkNotNull(obj1.toString());
  }
}
