/*
 * Copyright 2012 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.testdata;

public class MultipleTopLevelClassesWithNoErrors {
  int foo;
  int bar;

  public MultipleTopLevelClassesWithNoErrors(int foo, int bar) {
    this.foo = foo;
    this.bar = bar;
  }

  private static class InnerFoo {}
}

final class Foo1 {
  int foo;
  int bar;

  public Foo1(int foo, int bar) {
    this.foo = foo;
    this.bar = bar;
  }
}

final class Foo2 {
  int foo;
  int bar;

  public Foo2(int foo, int bar) {
    this.foo = foo;
    this.bar = bar;
  }
}

final class Foo3 {
  int foo;
  int bar;

  public Foo3(int foo, int bar) {
    this.foo = foo;
    this.bar = bar;
  }
}
