/*
 * Copyright 2014 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.threadsafety;

import com.google.errorprone.annotations.FormatMethod;
import com.google.errorprone.annotations.FormatString;

/**
 * An error that occurred during the parsing or binding of a GuardedBy expression.
 *
 * @author cushon@google.com (Liam Miller-Cushon)
 */
public class IllegalGuardedBy extends RuntimeException {
  public IllegalGuardedBy(String message) {
    super(message);
  }

  /** Throws an {@link IllegalGuardedBy} exception if the given condition is false. */
  public static void checkGuardedBy(boolean condition, String message) {
    if (!condition) {
      throw new IllegalGuardedBy(message);
    }
  }

  /** Throws an {@link IllegalGuardedBy} exception if the given condition is false. */
  @FormatMethod
  public static void checkGuardedBy(
      boolean condition, @FormatString String formatString, Object... formatArgs) {
    if (!condition) {
      throw new IllegalGuardedBy(String.format(formatString, formatArgs));
    }
  }

  @Override
  public String toString() {
    return getMessage();
  }
}
