/*
 * Copyright 2011 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone;

import com.google.errorprone.matchers.Description;
import com.sun.tools.javac.tree.JCTree.JCCompilationUnit;
import com.sun.tools.javac.util.Log;

/**
 * Strategies for reporting results.
 *
 * @author alexeagle@google.com (Alex Eagle)
 */
public interface DescriptionListener {

  /** Reports a suggested modification to the code. */
  void onDescribed(Description description);

  /** Factory for creating DescriptionListeners while compiling each file. */
  interface Factory {
    DescriptionListener getDescriptionListener(Log log, JCCompilationUnit compilation);
  }
}
