/*
 * Copyright 2011 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.apply;

import java.util.logging.Logger;

/** File destination which simply throws away the generated file. */
public class DiscardingFileDestination implements FileDestination {

  private static final Logger log = Logger.getLogger(DiscardingFileDestination.class.toString());

  @Override
  public void writeFile(SourceFile file) {
    log.info(String.format("Altered file %s thrown away", file.getPath()));
  }

  @Override
  public void flush() {}
}
