# Bug patterns

[TOC]

This list is auto-generated from our sources. Each bug pattern includes code
examples of both positive and negative cases; these examples are used in our
regression test suite.

Patterns which are marked __Experimental__ will not be evaluated against your
code, unless you specifically configure Error Prone. The default checks are
marked __On by default__, and each release promotes some experimental checks
after we've vetted them against Google's codebase.

{{#bugpatterns}}
## {{category}}

{{#checks}}
__[{{name}}](bugpattern/{{name}}.md)__ \
{{summary}}

{{/checks}}
{{/bugpatterns}}
