/* Copyright 2016 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.testdata;

import java.util.Comparator;

/**
 * @author sulku@google.com (Marsela Sulku)
 * @author mariasam@google.com (Maria Sam)
 */
public class ComparableAndComparatorPositiveCases {

  /** implements both interfaces */
  // BUG: Diagnostic contains: Class should not implement both
  public static class BadClass implements Comparable<BadClass>, Comparator<BadClass> {
    @Override
    public int compareTo(BadClass comparableNode) {
      return 0;
    }

    @Override
    public int compare(BadClass a, BadClass b) {
      return 0;
    }
  }

  /** Superclass test class */
  public static class SuperClass implements Comparator<SuperClass> {
    @Override
    public int compare(SuperClass o1, SuperClass o2) {
      return 0;
    }
  }

  /** SubClass test class */
  // BUG: Diagnostic contains: Class should not implement both
  public static class SubClass extends SuperClass implements Comparable<SubClass> {
    @Override
    public int compareTo(SubClass o) {
      return 0;
    }
  }
}
