/*
 * Copyright 2017 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.overloading.testdata;

public abstract class InconsistentOverloadsPositiveCasesVarargs {

  public void foo(String... rest) {}

  public void foo(int x, String... rest) {}

  // BUG: Diagnostic contains: foo(int x, int y, String... rest)
  public void foo(int y, int x, String... rest) {}

  abstract void bar(float x, float y);

  // BUG: Diagnostic contains: bar(float x, float y, float z, Object... rest)
  abstract void bar(float z, float y, float x, Object... rest);

  // BUG: Diagnostic contains: bar(float x, float y, float z, String string)
  abstract void bar(float y, String string, float x, float z);

  abstract void bar(Object... rest);
}
