/*
 * Copyright 2018 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.testdata;

/** @author bhagwani@google.com (Sumit Bhagwani) */
public class ObjectToStringPositiveCases {

  public static final class FinalObjectClassWithoutToString {}

  public static final class FinalGenericClassWithoutToString<T> {}

  void directToStringCalls() {
    FinalObjectClassWithoutToString finalObjectClassWithoutToString =
        new FinalObjectClassWithoutToString();
    // BUG: Diagnostic contains: ObjectToString
    System.out.println(finalObjectClassWithoutToString.toString());
  }

  void genericClassShowsErasure() {
    FinalGenericClassWithoutToString<Object> finalGenericClassWithoutToString =
        new FinalGenericClassWithoutToString<>();
    // BUG: Diagnostic contains: `FinalGenericClassWithoutToString@
    System.out.println(finalGenericClassWithoutToString.toString());
  }
}
