/*
 * Copyright 2016 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package com.google.errorprone.bugpatterns.testdata;

import static com.google.common.base.Throwables.propagateIfPossible;
import static com.google.common.base.Throwables.throwIfUnchecked;

import java.io.IOException;
import java.util.concurrent.ExecutionException;

/** @author cpovirk@google.com (Chris Povirk) */
public class ThrowIfUncheckedKnownCheckedTestPositiveCases {
  void simple(IOException e) {
    // BUG: Diagnostic contains: no-op
    throwIfUnchecked(e);

    // BUG: Diagnostic contains: no-op
    propagateIfPossible(e);
  }

  void union() {
    try {
      foo();
    } catch (IOException | ExecutionException e) {
      // BUG: Diagnostic contains: no-op
      throwIfUnchecked(e);

      // BUG: Diagnostic contains: no-op
      propagateIfPossible(e);
    }
  }

  <E extends IOException> void checkedGeneric(E e) {
    // BUG: Diagnostic contains: no-op
    throwIfUnchecked(e);
  }

  void foo() throws IOException, ExecutionException {}
}
