/*
 * Copyright 2018 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

syntax = "proto3";

package devtools.staticanalysis.errorprone.apidiff;

option java_package = "com.google.errorprone.bugpatterns.apidiff";

option java_outer_classname = "ApiDiffProto";

message Diff {
  repeated ClassDiff class_diff = 1;
}

message ClassDiff {
  oneof diff {
    EverythingDiff everything_diff = 1;
    MemberDiff member_diff = 2;
  }
}

message EverythingDiff {
  string class_name = 1;
}

message MemberDiff {
  string class_name = 1;
  repeated ClassMember member = 2;
}

message ClassMember {
  string identifier = 1;
  // a JVMS 4.3 member descriptor
  string member_descriptor = 2;
}