/*
 * Copyright 2016 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Bug patterns related to <a href="http://en.wikipedia.org/wiki/Dependency_injection">dependency
 * injection</a> and <a href="https://jcp.org/en/jsr/detail?id=330">JSR 330</a>. See the various
 * subpackages for checks related to specific DI frameworks.
 */
package com.google.errorprone.bugpatterns.inject;
