/*
 * Copyright 2013 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.refaster;

import com.google.common.testing.EqualsTester;
import com.google.common.testing.SerializableTester;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/**
 * Tests for {@link UConditional}.
 *
 * @author lowasser@google.com (Louis Wasserman)
 */
@RunWith(JUnit4.class)
public class UConditionalTest extends AbstractUTreeTest {

  @Test
  public void match() {
    assertUnifies(
        "true ? -1 : 1",
        UConditional.create(ULiteral.booleanLit(true), ULiteral.intLit(-1), ULiteral.intLit(1)));
  }

  @Test
  public void inline() {
    assertInlines(
        "true ? -1 : 1",
        UConditional.create(ULiteral.booleanLit(true), ULiteral.intLit(-1), ULiteral.intLit(1)));
  }

  @Test
  public void equality() {
    ULiteral trueLit = ULiteral.booleanLit(true);
    ULiteral falseLit = ULiteral.booleanLit(false);
    ULiteral negOneLit = ULiteral.intLit(-1);
    ULiteral oneLit = ULiteral.intLit(1);
    new EqualsTester()
        .addEqualityGroup(UConditional.create(trueLit, negOneLit, oneLit))
        .addEqualityGroup(UConditional.create(trueLit, oneLit, oneLit))
        .addEqualityGroup(UConditional.create(trueLit, negOneLit, negOneLit))
        .addEqualityGroup(UConditional.create(falseLit, negOneLit, oneLit))
        .testEquals();
  }

  @Test
  public void serialization() {
    SerializableTester.reserializeAndAssert(
        UConditional.create(ULiteral.booleanLit(true), ULiteral.intLit(-1), ULiteral.intLit(1)));
  }
}
