/*
 * Copyright 2012 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.matchers;

import com.google.errorprone.VisitorState;
import com.sun.source.tree.MethodTree;
import com.sun.source.tree.VariableTree;

/**
 * Matches if the given matcher matches all of/any of the parameters to this method.
 *
 * @author eaftan@google.com (Eddie Aftandilian)
 */
public class MethodHasParameters extends ChildMultiMatcher<MethodTree, VariableTree> {

  public MethodHasParameters(MatchType matchType, Matcher<VariableTree> nodeMatcher) {
    super(matchType, nodeMatcher);
  }

  @Override
  protected Iterable<? extends VariableTree> getChildNodes(
      MethodTree methodTree, VisitorState state) {
    return methodTree.getParameters();
  }
}
