/*
 * Copyright 2013 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.testdata;

/** @author scottjohnson@google.com (Scott Johnson) */
public class WrongParameterPackageNegativeCases {

  public void testParameter(Integer x) {}

  public void testParameter(Integer x, Integer y) {}

  public void testParameter2(Integer x, Integer y) {}

  /** Test overrides */
  public static class Subclass extends WrongParameterPackageNegativeCases {

    @Override
    public void testParameter(Integer x) {}

    @Override
    public void testParameter(Integer x, Integer y) {}

    public void testParameter(Boolean x, Integer y) {}

    public void testParameter(Boolean x) {}

    @Override
    public void testParameter2(WrongParameterPackageNegativeCases.Integer x, Integer y) {}

    @SuppressWarnings("ParameterPackage")
    public void testParameter(java.lang.Integer x) {}
  }

  /** Ambiguous Integer class */
  public static class Integer {}
}
