/*
 * Copyright 2012 The Error Prone Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.errorprone.bugpatterns.testdata;

/** @author alexloh@google.com (Alex Loh) */
public class ClassCanBeStaticPositiveCase2 {

  int outerVar1;
  int outerVar2;

  // Outer variable overridden
  // BUG: Diagnostic contains: private /* COMMENT */ static final class Inner2
  private /* COMMENT */ final class Inner2 {
    int outerVar1;
    int innerVar = outerVar1;

    int localMethod(int outerVar2) {
      return outerVar2;
    }
  }
}
